// This stays here because otherwise it's hard to tell if there's a parsing error
dump("###################################### frame-content loaded\n");

// how many milliseconds before the mousedown and the overlay of an element
const kTapOverlayTimeout = 200;

let Cc = Components.classes;
let Ci = Components.interfaces;
let gIOService = Cc["@mozilla.org/network/io-service;1"]
  .getService(Ci.nsIIOService);
let gFocusManager = Cc["@mozilla.org/focus-manager;1"]
  .getService(Ci.nsIFocusManager);

let HTMLIFrameElement = Ci.nsIDOMHTMLIFrameElement;
let HTMLFrameElement = Ci.nsIDOMHTMLFrameElement;
let HTMLTextAreaElement = Ci.nsIDOMHTMLTextAreaElement;
let HTMLInputElement = Ci.nsIDOMHTMLInputElement;
let HTMLSelectElement = Ci.nsIDOMHTMLSelectElement;
let HTMLLabelElement = Ci.nsIDOMHTMLLabelElement;
let HTMLButtonElement = Ci.nsIDOMHTMLButtonElement;
let HTMLOptGroupElement = Ci.nsIDOMHTMLOptGroupElement;
let HTMLOptionElement = Ci.nsIDOMHTMLOptionElement;

/** Send message to UI thread with browser guid as the first parameter. */
function sendMessage(name) {
  sendSyncMessage(name, Array.prototype.slice.call(arguments, 1));
}


// XXX get these in utils.js (part of 551258 refactor)

function getWindowUtils(win) {
  return win.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindowUtils);
}

function getScrollOffset(win) {
  var cwu = getWindowUtils(win);
  var scrollX = {};
  var scrollY = {};
  cwu.getScrollXY(false, scrollX, scrollY);
  return { x: scrollX.value, y: scrollY.value };
}

/**
 * @param x,y Browser coordinates
 * @return Element at position, null if no active browser or no element found
 */
function elementFromPoint(x, y) {

  // browser's elementFromPoint expect browser-relative client coordinates.
  // subtract browser's scroll values to adjust
  let cwu = getWindowUtils(content.document.defaultView);
  let scroll = getScrollOffset(content.document.defaultView);
  x = x - scroll.x;
  y = y - scroll.y;
  let elem = cwu.elementFromPoint(x, y,
                                  true,   /* ignore root scroll frame*/
                                  false); /* don't flush layout */

  // step through layers of IFRAMEs and FRAMES to find innermost element
  while (elem && (elem instanceof HTMLIFrameElement || elem instanceof HTMLFrameElement)) {
    // adjust client coordinates' origin to be top left of iframe viewport
    let rect = elem.getBoundingClientRect();
    x = x - rect.left;
    y = y - rect.top;
    elem = elem.contentDocument.elementFromPoint(x, y);
  }

  return elem;
}


function getBoundingContentRect(contentElem) {
  let window = content.document.defaultView;
  let offset = getScrollOffset(window);
  let r = contentElem.getBoundingClientRect();

  // step out of iframes and frames, offsetting scroll values
  try {
  for (let frame = contentElem.ownerDocument.defaultView; frame != window; frame = frame.parent) {
    // adjust client coordinates' origin to be top left of iframe viewport
    let rect = frame.frameElement.getBoundingClientRect();
    offset.add(rect.left, rect.top);
  }
  } catch (ex) { debugger; }

  return new Rect(r.left + offset.x, r.top + offset.y, r.width, r.height);
}




/** Reponsible for sending messages about viewport size changes and painting. */
function Coalescer() {
  this._pendingDirtyRect = new Rect(0, 0, 0, 0);
  this._pendingSizeChange = null;
  this._timer = new Util.Timeout(this);
  // XXX When moving back and forward in docShell history, MozAfterPaint does not get called properly and
  // some dirty rectangles are never flagged properly.  To fix this, coalescer will fake a paint event that
  // dirties the entire viewport.
  this._incremental = false;
}

Coalescer.prototype = {
  notify: function notify() {
    this.flush();
  },

  handleEvent: function handleEvent(e) {
    if (e.type == "MozAfterPaint") {
      var win = e.originalTarget;
      var scrollOffset = getScrollOffset(win);
      this.dirty(scrollOffset, e.clientRects);
    } else if (e.type == "MozScrolledAreaChanged") {
      // XXX if it's possible to get a scroll area change with the same values,
      // it would be optimal if this didn't send the same message twice.
      var doc = e.originalTarget;
      var win = doc.defaultView;
      var scrollOffset = getScrollOffset(win);
      if (win.parent != win)
	// We are only interested in root scroll pane changes
	return;
      this.sizeChange(scrollOffset, e.x, e.y, e.width, e.height);
    }
  },

  /** Next scroll size change event will invalidate all previous content. See constructor. */
  emptyPage: function emptyPage() {
    this._incremental = false;
  },

  startCoalescing: function startCoalescing() {
    this._timer.interval(1000);
  },
  
  stopCoalescing: function stopCoalescing() {
    this._timer.flush();
  },

  sizeChange: function sizeChange(scrollOffset, x, y, width, height) {
    // Adjust width and height from the incoming event properties so that we
    // ignore changes to width and height contributed by growth in page
    // quadrants other than x > 0 && y > 0.
    var x = x + scrollOffset.x;
    var y = y + scrollOffset.y;
    this._pendingSizeChange = {
      width: width + (x < 0 ? x : 0),
      height: height + (y < 0 ? y : 0)
    };

    // Clear any pending dirty rectangles since entire viewport will be invalidated
    // anyways.
    var rect = this._pendingDirtyRect;
    rect.top = rect.bottom;
    rect.left = rect.right;

    if (!this._timer.isPending())
      this.flush()
  },

  dirty: function dirty(scrollOffset, clientRects) {
    if (!this._pendingSizeChange) {
      var unionRect = this._pendingDirtyRect;
      for (var i = clientRects.length - 1; i >= 0; i--) {
        var e = clientRects.item(i);
        unionRect.expandToContain(new Rect(
          e.left + scrollOffset.x, e.top + scrollOffset.y, e.width, e.height));
      }

      if (!this._timer.isPending())
        this.flush()
    }
  },

  flush: function flush() {
    var dirtyRect = this._pendingDirtyRect;
    var sizeChange = this._pendingSizeChange;
    if (sizeChange) {
      sendMessage("FennecMozScrolledAreaChanged", sizeChange.width, sizeChange.height);
      if (!this._incremental)
        sendMessage("FennecMozAfterPaint", [new Rect(0, 0, sizeChange.width, sizeChange.height)]);
      this._pendingSizeChange = null;
      // After first size change has been issued, assume subsequent size changes are only incremental
      // changes to the current page.
      this._incremental = true;
    }
    else if (!dirtyRect.isEmpty()) {
      // No size change has occurred, but areas have been dirtied.
      sendMessage("FennecMozAfterPaint", [dirtyRect]);
      dirtyRect.top = dirtyRect.bottom;
      dirtyRect.left = dirtyRect.right;
    }
  }
};


/**
 * Responsible for sending messages about security, location, and page load state.
 * @param loadingController Object with methods startLoading and stopLoading
 */
function ProgressController(loadingController) {
  this._webNavigation = docShell.QueryInterface(Ci.nsIWebNavigation);
  this._overrideService = null;
  this._hostChanged = false;
  this._state = null;
  this._loadingController = loadingController || this._defaultLoadingController;
}

ProgressController.prototype = {
  // Default loading callbacks do nothing
  _defaultLoadingController: {
    startLoading: function() {},
    stopLoading: function() {}
  },

  /**
   * Helper to parse out the important parts of the SSL cert for use in constructing
   * identity UI strings
   */
  _getIdentityData: function() {
    var result = {};
    var status = docShell.securityUI
      .QueryInterface(Ci.nsISSLStatusProvider)
      .SSLStatus;

    if (status) {
      status = status.QueryInterface(Ci.nsISSLStatus);
      var cert = status.serverCert;

      // Human readable name of Subject
      result.subjectOrg = cert.organization;

      // SubjectName fields, broken up for individual access
      if (cert.subjectName) {
        result.subjectNameFields = {};
        cert.subjectName.split(",").forEach(function(v) {
          var field = v.split("=");
          if (field[1])
            this[field[0]] = field[1];
        }, result.subjectNameFields);

        // Call out city, state, and country specifically
        result.city = result.subjectNameFields.L;
        result.state = result.subjectNameFields.ST;
        result.country = result.subjectNameFields.C;
      }

      // Human readable name of Certificate Authority
      result.caOrg =  cert.issuerOrganization || cert.issuerCommonName;

      if (!this._overrideService)
        this._overrideService = Cc["@mozilla.org/security/certoverride;1"].getService(Ci.nsICertOverrideService);

      // Check whether this site is a security exception. XPConnect does the right
      // thing here in terms of converting _lastLocation.port from string to int, but
      // the overrideService doesn't like undefined ports, so make sure we have
      // something in the default case (bug 432241).
      result.isException = !!this._overrideService.hasMatchingOverride(content.location.hostname,
        (content.location.port || 443), cert, {}, {});
    }

    return result;
  },

  onStateChange: function onStateChange(aWebProgress, aRequest, aStateFlags, aStatus) {
    // ignore notification that aren't about the main document (iframes, etc)
    var win = aWebProgress.DOMWindow;
    if (win != win.parent)
      return;

    // If you want to observe other state flags, be sure they're listed in the
    // Tab._createBrowser's call to addProgressListener
    if (aStateFlags & Ci.nsIWebProgressListener.STATE_IS_NETWORK) {
      if (aStateFlags & Ci.nsIWebProgressListener.STATE_START) {
        this._loadingController.startLoading();
        sendMessage("FennecNetworkStart");
      }
      else if (aStateFlags & Ci.nsIWebProgressListener.STATE_STOP) {
        this._loadingController.stopLoading();
        sendMessage("FennecNetworkStop");
      }
    }
    else if (aStateFlags & Ci.nsIWebProgressListener.STATE_IS_DOCUMENT) {
      if (aStateFlags & Ci.nsIWebProgressListener.STATE_STOP) {
        sendMessage("FennecDocumentStop");
      }
    }
  },

  /** This method is called to indicate progress changes for the currently loading page. */
  onProgressChange: function onProgressChange(aWebProgress, aRequest, aCurSelf, aMaxSelf, aCurTotal, aMaxTotal) {
    // To use this method, add NOTIFY_PROGRESS to the flags
  },

  /** This method is called to indicate a change to the current location. */
  onLocationChange: function onLocationChange(aWebProgress, aRequest, aLocationURI) {
    var location = aLocationURI ? aLocationURI.spec : "";
    this._hostChanged = true;
    sendMessage("FennecLocationChange", location);
    sendMessage("FennecDocshellHistory", docShell.canGoBack, docShell.canGoForward);
  },

  /**
   * This method is called to indicate a status changes for the currently
   * loading page.  The message is already formatted for display.
   */
  onStatusChange: function onStatusChange(aWebProgress, aRequest, aStatus, aMessage) {
    // To use this method, add NOTIFY_STATUS to the flags
  },

  /** This method is called when the security state of the browser changes. */
  onSecurityChange: function onSecurityChange(aWebProgress, aRequest, aState) {
    // Don't need to do anything if the data we use to update the UI hasn't changed
    if (this._state == aState && !this._hostChanged)
      return;

    this._hostChanged = false;
    this._state = aState;

    // It's important that this message is synchronous!
    sendMessage("FennecSecurityChange", aState, this._getIdentityData());
  },

  QueryInterface: function QueryInterface(aIID) {
    if (aIID.equals(Ci.nsIWebProgressListener) ||
        aIID.equals(Ci.nsISupportsWeakReference) ||
        aIID.equals(Ci.nsISupports)) {
        return this;
    }

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  start: function start() {
    let flags = Ci.nsIWebProgress.NOTIFY_LOCATION |
      Ci.nsIWebProgress.NOTIFY_SECURITY |
      Ci.nsIWebProgress.NOTIFY_STATE_NETWORK |
      Ci.nsIWebProgress.NOTIFY_STATE_DOCUMENT;

    let webProgress = docShell.QueryInterface(Ci.nsIInterfaceRequestor)
      .getInterface(Ci.nsIWebProgress);
    webProgress.addProgressListener(this, flags);
  },

  stop: function stop() {
    let webProgress = docShell.QueryInterface(Ci.nsIInterfaceRequestor)
      .getInterface(Ci.nsIWebProgress);
    webProgress.removeProgressListener(this);
  }
};


/** Responsible for receiving Fennec messages that tell the docShell to do something. */
function BrowserDelegate() {
  this._webNavigation = docShell.QueryInterface(Ci.nsIWebNavigation);

  addMessageListener("FennecSearch", this);
  addMessageListener("FennecReload", this);
  addMessageListener("FennecStop", this);
  addMessageListener("FennecBack", this);
  addMessageListener("FennecForward", this);
}

BrowserDelegate.prototype = {
  receiveMessage: Util.receiveMessage,

  receiveFennecReload: function(flags) {
    flags = flags || this._webNavigation.LOAD_FLAGS_NONE;
    this._webNavigation.reload(flags);
  },

  receiveFennecStop: function() {
    let flags = this._webNavigation.STOP_ALL;
    this._webNavigation.stop(flags);
  },

  receiveFennecSearch: function receiveFennecSearch(message, uri, postData) {
    this._webNavigation.loadURI(uri, null, null, postData, null);
  },

  receiveFennecBack: function receiveFennecBack() {
    this._webNavigation.goBack();
  },

  receiveFennecForward: function receiveFennecForward() {
    this._webNavigation.goForward();
  }
};


/**
 * Responsible for opening up form assistant and sending messages to the FormHelper when user
 * types keys to navigate.
 */
function ContentFormManager() {
  this._navigator = null;

  addMessageListener("FennecClosedFormAssist", this);
  addMessageListener("FennecFormPrevious", this);
  addMessageListener("FennecFormNext", this);
  addMessageListener("FennecFormChoiceSelect", this);
  addMessageListener("FennecFormChoiceChange", this);
}

ContentFormManager.prototype = {
  formAssist: function(element) {
    let wrapper = new BasicWrapper(element);
    if (!wrapper.canAssist())
      return false;

    let navigator = new FormNavigator(this, content.document);
    if (!navigator.setCurrent(element))
      return false;

    sendMessage("FennecFormAssist", navigator.getJSON());

    if (!this._navigator)
      addEventListener("keyup", this, false);
    this._navigator = navigator;
    return true;
  },

  closeFormAssistant: function() {
    if (this._navigator) {
      sendMessage("FennecCloseFormAssist");
      this.closedFormAssistant();
    }
  },

  closedFormAssistant: function() {
    if (this._navigator) {
      removeEventListener("keyup", this, false);
      this._navigator = null;
    }
  },

  receiveMessage: Util.receiveMessage,

  receiveFennecClosedFormAssist: function() {
    this.closedFormAssistant();
  },

  receiveFennecFormPrevious: function() {
    if (this._navigator) {
      this._navigator.goToPrevious();
      sendMessage("FennecFormAssist", this._navigator.getJSON());
    }
  },

  receiveFennecFormNext: function() {
    if (this._navigator) {
      this._navigator.goToNext();
      sendMessage("FennecFormAssist", this._navigator.getJSON());
    }
  },

  receiveFennecFormChoiceSelect: function(message, index, selected, clearAll) {
    if (this._navigator) {
      let current = this._navigator.getCurrent();
      if (current)
        current.choiceSelect(index, selected, clearAll);
    }
  },

  receiveFennecFormChoiceChange: function() {
    if (this._navigator) {
      let current = this._navigator.getCurrent();
      if (current)
        current.choiceChange();
    }
  },

  handleEvent: function formHelperHandleEvent(aEvent) {
    let currentWrapper = this._navigator.getCurrent();
    let currentElement = currentWrapper.element;

    switch (aEvent.keyCode) {
      case aEvent.DOM_VK_DOWN:
        if (currentElement instanceof HTMLTextAreaElement) {
          let existSelection = currentElement.selectionEnd - currentElement.selectionStart;
          let isEnd = (currentElement.textLength == currentElement.selectionEnd);
          if (!isEnd || existSelection)
            return;
        }

        this._navigator.goToNext();
        break;

      case aEvent.DOM_VK_UP:
        if (currentElement instanceof HTMLTextAreaElement) {
          let existSelection = currentElement.selectionEnd - currentElement.selectionStart;
          let isStart = (currentElement.selectionEnd == 0);
          if (!isStart || existSelection)
            return;
        }

        this._navigator.goToPrevious();
        break;

      case aEvent.DOM_VK_RETURN:
        break;

      default:
        let target = aEvent.target;
        if (currentWrapper.canAutocomplete())
          sendMessage("FennecFormAutocomplete", currentWrapper.getAutocompleteSuggestions());
        break;
    }
  }
};


/**
 * Responsible for iterating through form elements.
 * The navigable list is generated on instantiation, so construct new FormNavigators when
 * the current document is changed or the list of form elements needs to be regenerated.
 */
function FormNavigator(manager, document) {
  this._manager = manager;
  this._wrappers = this._getAllWrappers(document);
  this._currentIndex = 0;
}

FormNavigator.prototype = {
  endNavigation: function() {
    this._manager.closedFormAssistant();
  },

  getCurrent: function() {
    return this._wrappers[this._currentIndex];
  },

  getJSON: function() {
    return {
      hasNext: !!this.getNext(),
      hasPrevious: !!this.getPrevious(),
      current: this.getCurrent().getJSON()
    };
  },

  // XXX doesn't really belong.
  setCurrent: function(element) {
    for (var i = 0; i < this._wrappers.length; i++) {
      let node = this._wrappers[i];
      if (node.element == element) {
        return this._setIndex(i);
      }
    }
    return false;
  },

  getPrevious: function() {
    return this._wrappers[this._currentIndex - 1];
  },

  getNext: function() {
    return this._wrappers[this._currentIndex + 1];
  },

  goToPrevious: function() {
    return this._setIndex(this._currentIndex - 1);
  },

  goToNext: function() {
    return this._setIndex(this._currentIndex + 1);
  },

  _getAllWrappers: function(document) {
    // XXX good candidate for tracing if possible.  The tough ones are length and
    // canNavigateTo / isVisible.

    if (!document)
      return [];

    let elements = [];

    // get all the documents
    let documents = [document];
    let iframes = document.querySelectorAll("iframe, frame");
    for (let i = 0; i < iframes.length; i++)
      documents.push(iframes[i].contentDocument);

    for (let i = 0; i < documents.length; i++) {
      let nodes = documents[i].querySelectorAll("input, button, select, textarea, [role=button]");

      for (let j = 0; j < nodes.length; j++) {
        let node = nodes[j];
        let wrapper = new BasicWrapper(node);
        if (wrapper.canNavigateTo() && wrapper.isVisible())
          elements.push(wrapper);
      }
    }

    function orderByTabIndex(a, b) {
      // for an explanation on tabbing navigation see
      // http://www.w3.org/TR/html401/interact/forms.html#h-17.11.1
      // In resume tab index navigation order is 1, 2, 3, ..., 32767, 0
      if (a.tabIndex == 0 || b.tabIndex == 0)
        return b.tabIndex;

      return a.tabIndex > b.tabIndex;
    }
    return elements.sort(orderByTabIndex);
  },

  _setIndex: function(i) {
    let element = this._wrappers[i];
    if (element) {
      gFocusManager.setFocus(element.element, Ci.nsIFocusManager.FLAG_NOSCROLL);
      this._currentIndex = i;
      return true;
    } else {
      return false;
    }
  }
};


/** Can't think of a good description of this class.  It probably does too much? */
function Content() {
  this._iconURI = null;

  addEventListener("DOMTitleChanged", this, false);
  addEventListener("DOMLinkAdded", this, false);
  addMessageListener("FennecBlur", this);
  addMessageListener("FennecFocus", this);
  addMessageListener("FennecMousedown", this);
  addMessageListener("FennecMouseup", this);
  addMessageListener("FennecCancelMouse", this);

  this._coalescer = new Coalescer();
  addEventListener("MozAfterPaint", this._coalescer, false);
  addEventListener("MozScrolledAreaChanged", this._coalescer, false);

  this._progressController = new ProgressController(this);
  this._progressController.start();

  this._browserDelegate = new BrowserDelegate();
  this._contentFormManager = new ContentFormManager();

  this._mousedownTimeout = new Util.Timeout();
}

Content.prototype = {
  handleEvent: Util.handleEvent,

  receiveMessage: Util.receiveMessage,

  handleDOMWindowClose: function(e) {
    if (e.isTrusted)
      sendMessage("FennecClose");
  },

  handleDOMWillOpenModalDialog: function(e) {
    if (e.isTrusted)
      sendMessage("FennecWillOpenModalDialog");
  },

  handleDOMTitleChanged: function(e) {
    sendMessage("FennecTitle", content.document.title);
  },

  handleDOMLinkAdded: function(e) {
    let link = e.originalTarget;
    let ownerDoc = link.ownerDocument;
    // Only check links that are in the root document.
    if (!link || !link.href || !ownerDoc || ownerDoc.defaultView.frameElement)
      return;

    if (/\bicon\b/i(link.rel)) {
      // Link added is a favicon
      this.setIcon(link.href);
    }
    else if (/\bsearch\b/i(link.rel) && link.type && link.title) {
      // Link added is a search engine
      let type = link.type.replace(/^\s+|\s*(?:;.*)?$/g, "").toLowerCase();
      if (type == "application/opensearchdescription+xml" && /^(?:https?|ftp):/i.test(link.href)) {
        let engine = { title: link.title, href: link.href };
        sendMessage("FennecAddSearch", engine);
      }
    }
  },

  receiveFennecFocus: function receiveFennecFocus() {
    docShell.isOffScreenBrowser = true;
    this._selected = true;
  },

  receiveFennecBlur: function receiveFennecBlur() {
    docShell.isOffScreenBrowser = false;
    this._selected = false;
  },

  _sendMouseEvent: function _sendMouseEvent(name, x, y) {
    let window = content.document.defaultView;
    let windowUtils = getWindowUtils(window);
    let scrollOffset = getScrollOffset(window);
    windowUtils.sendMouseEvent(name, x - scrollOffset.x, y - scrollOffset.y, 0, 1, 0, true);
  },

  receiveFennecMousedown: function receiveFennecMousedown(message, x, y) {
    this._mousedownTimeout.once(kTapOverlayTimeout, function() {
      let element = elementFromPoint(x, y);
      gFocusManager.setFocus(element, Ci.nsIFocusManager.FLAG_NOSCROLL);
    });
  },

  receiveFennecMouseup: function receiveFennecMouseup(message, x, y) {
    this._mousedownTimeout.flush();

    let element = elementFromPoint(x, y);
    Util.dumpLn('x and y', x, y);
    if (!this._contentFormManager.formAssist(element)) {
      this._sendMouseEvent("mousedown", x, y);
      this._sendMouseEvent("mouseup", x, y);
    }
  },

  receiveFennecCancelMouse: function receiveFennecCancelMouse() {
    this._mousedownTimeout.clear();
    // XXX there must be a better way than this to cancel the mouseover/focus?
    this._sendMouseEvent("mouseup", -1000, -1000);
    let element = content.document.activeElement;
    if (element)
      element.blur();
  },

  setIcon: function setIcon(uriSpec) {
    let uri = null;
    try {
      if (uriSpec)
        uri = gIOService.newURI(uriSpec,
          content.document.characterSet, content.document.documentURIObject);
    } catch(e) {
    }

    if (uri && !uri.schemeIs("javascript")) {
      this._iconURI = uri.spec;
    }

    if (!this._loading) {
      let defaultIconURI = content.document.documentURIObject.prePath + "/favicon.ico";
      sendMessage("FennecIcon", this._iconURI, defaultIconURI);
    }
  },

  startLoading: function startLoading() {
    this._loading = true;
    this._iconURI = null;
    this._coalescer.emptyPage();
    this._coalescer.startCoalescing();
  },

  stopLoading: function stopLoading() {
    this._loading = false;
    this._coalescer.stopCoalescing();
    this.setIcon();
  },

  isSelected: function isSelected() {
    return this._selected;
  }
};


let contentObject = new Content();
